from __future__ import absolute_import, unicode_literals

from thorn.django.models import Subscriber

from django.contrib.auth import get_user_model

from thorn.tests.case import Case


class test_SubscriberManager(Case):

    def setUp(self):
        self.user, _ = get_user_model().objects.get_or_create(
            username='test_Subscriber',
            password='iesf&83j2aswe2',
            email='example@example.com',
        )
        self.user2, _ = get_user_model().objects.get_or_create(
            username='test_Subscriber2',
            password='asdjqwej21j2344',
            email='example2@example.com',
        )
        try:
            self.subscribers = [
                self.rsimple('foo.*', 'A'),
                self.rsimple('foo.created', 'B'),
                self.rsimple('foo.created', 'C'),
                self.rsimple('foo.deleted', 'D'),
                self.rsimple('*.created', 'E'),
                self.rsimple('bar.updated', 'F'),
                self.rsimple('baz.*', 'G', user=self.user2),
            ]
        except BaseException:  # pragma: no cover
            raise

    def tearDown(self):
        self.user.delete()
        [r.delete() for r in self.subscribers]

    def rsimple(self, event, url, user=None):
        return Subscriber.objects.create(
            event=event, url=url, user=user or self.user,
        )

    def test_matching(self):
        self.assertFalse(
            Subscriber.objects.matching('baz.moo', user=self.user)
        )
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching(
                'baz.moo', user=self.user2)],
            ['G'],
        )

    def test_matching_event(self):
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('bar.updated')],
            ['F'],
        )

    def test_matching_event__glob(self):
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('foo.created')],
            ['A', 'B', 'C', 'E'],
        )
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('foo.updated')],
            ['A'],
        )
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('foo.deleted')],
            ['A', 'D'],
        )
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('baz.created')],
            ['E', 'G'],
        )
        self.assertListEqual(
            [r.url for r in Subscriber.objects.matching_event('bar.deleted')],
            [],
        )



"""

    thorn.funtests.base
    =================================

    Extends Cyanide stress test suite with utilities used
    to test Thorn.

"""
from __future__ import absolute_import, unicode_literals

import hashlib
import json
import requests

from contextlib import contextmanager
from six import iteritems as items, itervalues as values
from uuid import uuid4

from django.contrib.auth import get_user_model
from django.core.urlresolvers import reverse

from cyanide.suite import Suite, testcase
from itsdangerous import Signer

from thorn.django.models import Subscriber

from testapp.models import Article, SubscriberLog, Tag

BASE_URL = 'http://localhost:8000'

__all__ = ['WebhookSuite', 'new_ref', 'url', 'event_url', 'testcase']


def new_ref():
    return uuid4().hex


def url(*s):
    return '/'.join((BASE_URL,) + s)


def event_url(event, ref=None, rest=None):
    return url('r', event, '?ref={0}{1}'.format(ref, rest or ''))


class WebhookSuite(Suite):
    user = user2 = None
    token = None
    token_type = 'Token'

    def setup(self):
        Article.objects.all().delete()
        Tag.objects.all().delete()
        SubscriberLog.objects.all().delete()
        Subscriber.objects.all().delete()

        self.user = self._get_user('test', 'test')
        self.user2 = self._get_user('test2', 'test2')
        self.token = self._login()
        self.ref = new_ref()
        assert not SubscriberLog.objects.filter(ref=self.ref)

    def create_article(self, title, state='PENDING', author=None):
        return Article.objects.create(
            title=title, state=state, author=author or self.user,
        )

    def assert_webhook_not_received(self, ref=None):
        return self.ensure_not_for_a_while(
            SubscriberLog.objects.get, SubscriberLog.DoesNotExist,
            desc='webhook received [ref={0}]'.format(ref or self.ref),
            kwargs={'ref': ref or self.ref},
        )

    def subscribe(self, event, ref=None, rest=None):
        return self.post(
            'hooks/',
            event=event, url=self._event_url(event, ref, rest),
        )

    def unsubscribe(self, url):
        return self._delete(url)

    def list_subscriptions(self):
        return self.get('hooks/')

    def _event_url(self, event, ref=None, rest=None):
        return event_url(event, ref=ref or self.ref, rest=rest)

    def assert_article_event_received(self, article, event, sub=None,
                                      reverse=None, ref=None, n=1):
        logs = self.wait_for_webhook_received(ref or self.ref, maxlen=n)
        assert len(logs) == n
        if sub is not None:
            hmac_secret = sub['hmac_secret']
            if hmac_secret:
                log = SubscriberLog.objects.filter(ref=ref or self.ref)[0]
                assert Signer(
                    hmac_secret,
                    digest_method=hashlib.sha256).get_signature(
                        log.data) == log.hmac
        self.assert_log_matches(
            logs[0],
            event=event,
            ref=reverse or self.reverse_article(article),
            data=article.webhook_payload(),
        )

    def reverse_article(self, article):
        return reverse('article:detail', kwargs={'id': article.pk})

    def wait_for_webhook_received(self, ref=None, maxlen=1):
        return self.wait_for(
            self.assert_log, Exception,
            'webhook (ref={0})'.format(ref or self.ref),
            args=(ref or self.ref, maxlen),
        )

    def assert_log(self, ref=None, maxlen=1):
        logs = SubscriberLog.objects.filter(ref=ref or self.ref)
        assert logs and (len(logs) == maxlen if maxlen else True)
        return [json.loads(entry.data) for entry in logs]

    def assert_log_matches(self, log, **expected):
        for k, v in items(expected):
            assert log[k] == v, 'key={0} expected {1} == {2}'.format(
                k, v, log[k])

    def _login(self, username='test', password='test'):
        return self.post(
            'api-token-auth/',
            username=username,
            password=password,
        )['token']

    def _get_user(self, username, password, email='test@example.com'):
        try:
            return get_user_model().objects.create_user(
                username=username, password=password, email=email,
            )
        except Exception:
            return get_user_model().objects.get(username=username)

    def headers(self):
        if self.token:
            return {
                'Authorization': ' '.join([
                    self.token_type.capitalize(), self.token,
                ])
            }
        return {}

    @contextmanager
    def override_worker_setting(self, setting_name, new_value):
        old_value = list(values(self.setenv(setting_name, new_value)))[0]
        try:
            yield
        finally:
            self.setenv(setting_name, old_value)

    @contextmanager
    def worker_subscribe_to(self, event, url=None, callback=None):
        self.hook_subscribe(event, url=url, callback=callback)
        try:
            yield
        finally:
            self.hook_clear(event)

    def setenv(self, setting_name, new_value):
        return self.app.control.inspect()._request(
            'setenv', setting_name=setting_name, new_value=new_value)

    def assert_ok_pidbox_response(self, replies):
        for reply in values(replies):
            if not reply['ok']:
                raise RuntimeError(
                    'Worker remote control command raised: {0!r}'.format(
                        reply.get('error', reply)))
        return replies

    def hook_subscribe(self, event, url, callback=None):
        return self.assert_ok_pidbox_response(
            self.app.control.inspect()._request(
                'hook_subscribe', event=event, url=url, callback=callback,
            ),
        )

    def hook_unsubscribe(self, event, url):
        return self.assert_ok_pidbox_response(
            self.app.control.inspect()._request(
                'hook_unsubscribe', event=event, url=url,
            ),
        )

    def hook_clear(self, event):
        return self.assert_ok_pidbox_response(
            self.app.control.inspect()._request(
                'hook_clear', event=event,
            )
        )

    def get(self, *path, **data):
        return self._request(requests.get, 'data', url(*path), data)

    def post(self, *path, **data):
        return self._request(requests.post, 'json', url(*path), data)

    def delete(self, *path, **data):
        return self._request(requests.delete, 'json', url(*path), data)

    def _delete(self, url, **data):
        return self._request(requests.delete, 'json', url, data)

    def _request(self, fun, data_key, url, data):
        response = fun(url, headers=self.headers(), **{data_key: data})
        try:
            response.raise_for_status()
        except Exception:
            self.error('HTTP response body was: {0!r}'.format(
                response.content))
            raise
        return response.json() if response.content else None



from __future__ import absolute_import, unicode_literals

from six import iteritems as items

from celery import Celery, current_app

from django.db.models import signals

from thorn import Thorn
from thorn.events import Event
from thorn import _state

from nose import SkipTest

from case import (
    ANY, Case, MagicMock, Mock, patch, call, mock, skip,
)

__all__ = [
    'ANY', 'Case', 'MagicMock', 'Mock', 'SkipTest',
    'patch', 'call', 'mock', 'skip',

    'EventCase', 'RealSignalCase', 'SignalCase',
]

DEFAULT_RECIPIENT_VALIDATORS = [
    ('block_internal_ips', ()),
    ('ensure_protocol', ('http://', 'https://')),
    ('ensure_port', (80, 443)),
]


def _reset_signal(sig, receivers):
    receivers, sig.receivers = sig.receivers, receivers
    try:
        sig.sender_receivers_cache.clear()
    except AttributeError:  # pragma: no cover
        pass
    return receivers


class RealSignalCase(Case):

    reset_signals = {signals.post_save, signals.post_delete}

    def setUp(self):
        self.on_setup_track_signals()
        super(RealSignalCase, self).setUp()

    def tearDown(self):
        self.on_teardown_reset_signals()
        super(RealSignalCase, self).tearDown()

    def on_setup_track_signals(self):
        self._sigstate = {
            sig: _reset_signal(sig, []) for sig in self.reset_signals
        }

    def on_teardown_reset_signals(self):
        for signal, receivers in items(self._sigstate):
            _reset_signal(signal, receivers)


class ThornCase(Case):

    def setUp(self):
        self._tls, _state._tls = _state._tls, _state._TLS()
        self.app = Thorn(set_as_current=True)
        self._default_app, _state.default_app = _state.default_app, self.app
        super(ThornCase, self).setUp()

    def tearDown(self):
        _state.default_app = self._default_app
        _state._tls = self._tls
        super(ThornCase, self).tearDown()


class SignalCase(ThornCase):

    def setUp(self):
        self.Model = Mock(name='Model')
        self.post_save = self.patch('django.db.models.signals.post_save')
        self.post_delete = self.patch('django.db.models.signals.post_delete')
        self.m2m_changed = self.patch('django.db.models.signals.m2m_changed')
        super(SignalCase, self).setUp()


class EventCase(SignalCase):
    Dispatcher = Mock
    Event = Event

    def setUp(self):
        self._prev_app = current_app._get_current_object()
        self.app = Celery(set_as_current=True)
        self.dispatcher = self.Dispatcher()
        self.event = self.mock_event('george.costanza')
        super(EventCase, self).setUp()

    def mock_event(self, name, dispatcher=None, app=None, **kwargs):
        return self.Event(
            name,
            dispatcher=dispatcher or self.dispatcher,
            app=app or self.app,
            **kwargs
        )

    def tearDown(self):
        self._prev_app.set_current()
        super(EventCase, self).tearDown()



"""

    thorn.utils.compat
    ==================

    Python version compatibility utilities.

"""
from __future__ import absolute_import, unicode_literals

import random
import string
import sys

__all__ = ['bytes_if_py2']

PY3 = sys.version_info[0] >= 3

punctuation = string.punctuation.replace('"', '').replace("'", '')

if PY3:  # pragma: no cover
    def bytes_if_py2(s):
        return s
else:  # pragma: no cover
    def bytes_if_py2(s):  # noqa
        if isinstance(s, unicode):
            return s.encode()
        return s


def restore_from_keys(fun, args, kwargs):
    return fun(*args, **kwargs)


def random_secret(
        length, chars=string.ascii_letters + string.digits + punctuation):
    return ''.join(random.choice(chars) for _ in range(length))



"""

    thorn.django.rest_framework.serializers
    =======================================

    Rest framework serializers.

"""
from __future__ import absolute_import, unicode_literals

from rest_framework import serializers

from thorn.django.models import Subscriber

__all__ = ['SubscriberSerializer']


class SubscriberSerializer(serializers.HyperlinkedModelSerializer):
    subscription = serializers.HyperlinkedIdentityField(
        view_name='webhook:detail',
        lookup_url_kwarg='uuid', lookup_field='uuid',
    )
    id = serializers.UUIDField(source='uuid', read_only=True)
    user = serializers.IntegerField(source='user.pk', default=None)

    class Meta:
        model = Subscriber
        fields = (
            'event', 'url', 'content_type', 'user',
            'id', 'created_at', 'updated_at', 'subscription',
            'hmac_secret', 'hmac_digest',
        )
        read_only_fields = ('id', 'created_at', 'updated_at', 'subscription')



"""

    thorn.django.admin
    ==================

    Django-Admin interface support for managing webhook subscriptions.

"""
from __future__ import absolute_import, unicode_literals

from django.contrib import admin

from . import models

admin.site.register(models.Subscriber)



"""

    thorn.exceptions
    ================

    Thorn-related exceptions.

"""
from __future__ import absolute_import, unicode_literals


class ThornError(Exception):
    pass


class ImproperlyConfigured(ThornError):
    pass


class SecurityError(ThornError):
    pass



from __future__ import absolute_import, unicode_literals

from thorn.django.utils import serialize_model, deserialize_model

from testapp.models import Article

from thorn.tests.case import Case, Mock, patch


class test_deserialize_model(Case):

    def setup(self):
        self.Model = Mock(name='Model')
        self.Model._meta.app_label = 'app_label'
        self.Model._meta.model_name = 'model_name'

    def test_serialize(self):
        self.assertDictEqual(serialize_model(self.Model), {
            'app_label': self.Model._meta.app_label,
            'model_name': self.Model._meta.model_name,
        })

    @patch('thorn.django.utils.apps')
    def test_deserialize(self, django_apps):
        self.assertIs(
            deserialize_model(serialize_model(self.Model)),
            django_apps.get_model.return_value,
        )
        django_apps.get_model.assert_called_with(
            self.Model._meta.app_label,
            self.Model._meta.model_name,
        )

    def test_deserialize__functional_from_class(self):
        self.assertIs(deserialize_model(serialize_model(Article)), Article)

    def test_deserialize__functional_from_instance(self):
        instance = Article(title='foo')
        self.assertIs(deserialize_model(serialize_model(instance)), Article)

    def test_deserialize__already_model(self):
        self.assertIs(deserialize_model(Article), Article)



"""

    thorn.environment.django
    ========================

    Django web framework environment.

"""
from __future__ import absolute_import, unicode_literals

import importlib
import os

from celery.utils import cached_property
from celery.utils.imports import symbol_by_name

__all__ = ['DjangoEnv']


class DjangoEnv(object):
    settings_cls = 'django.conf:settings'
    subscriber_cls = 'thorn.django.models:Subscriber'
    signals_cls = 'thorn.django.signals'
    reverse_cls = 'django.core.urlresolvers:reverse'

    @staticmethod
    def autodetect(env='DJANGO_SETTINGS_MODULE'):
        return os.environ.get(env)

    @cached_property
    def config(self):
        return symbol_by_name(self.settings_cls)

    @cached_property
    def Subscriber(self):
        return symbol_by_name(self.subscriber_cls)

    @cached_property
    def Subscribers(self):
        return self.Subscriber.objects

    @cached_property
    def signals(self):
        return importlib.import_module(self.signals_cls)

    @cached_property
    def reverse(self):
        return symbol_by_name(self.reverse_cls)



from __future__ import absolute_import, unicode_literals

import pytz

from decimal import Decimal
from datetime import datetime
from json import loads
from six import text_type
from uuid import uuid4

from thorn.utils.json import JsonEncoder, dumps, get_best_json

from thorn.tests.case import Case, Mock, patch


class test_JsonEncoder(Case):

    def test_datetime(self):
        now = datetime.utcnow()
        now_utc = now.replace(tzinfo=pytz.utc)
        stripped = datetime(*now.timetuple()[:3])
        serialized = loads(dumps({
            'datetime': now,
            'tz': now_utc,
            'date': now.date(),
            'time': now.time()},
        ))
        self.assertDictEqual(serialized, {
            'datetime': now.isoformat(),
            'tz': '{0}Z'.format(now_utc.isoformat().split('+', 1)[0]),
            'time': now.time().isoformat(),
            'date': stripped.isoformat(),
        })

    def test_Decimal(self):
        d = Decimal('3314132.13363235235324234123213213214134')
        self.assertDictEqual(loads(dumps({'d': d})), {
            'd': text_type(d),
        })

    def test_UUID(self):
        id = uuid4()
        self.assertDictEqual(loads(dumps({'u': id})), {
            'u': text_type(id),
        })

    def test_default(self):
        with self.assertRaises(TypeError):
            dumps({'o': object()})


class test_dumps(Case):

    def test_dumps(self):
        obj = Mock(name='obj')
        encode = Mock(name='encode')
        self.assertIs(dumps(obj, encode=encode), encode.return_value)
        encode.assert_called_with(obj, cls=JsonEncoder)


class test_get_best_json(Case):

    @patch('thorn.utils.json.symbol_by_name')
    def test_no_alternatives(self, symbol_by_name):
        symbol_by_name.side_effect = ImportError()
        with self.assertRaises(ImportError):
            get_best_json()

    def test_no_choices(self):
        get_best_json(choices=[])



"""Python Webhook and Event Framework"""
from __future__ import absolute_import, unicode_literals

from .app import Thorn
from .events import Event, ModelEvent
from .decorators import webhook_model
from .reverse import model_reverser
from .utils.functional import Q


VERSION = (1, 2, 0)
__version__ = '.'.join(map(str, VERSION[0:3])) + ''.join(VERSION[3:])
__author__ = 'Robinhood Markets'
__contact__ = 'thorn@robinhood.com'
__homepage__ = 'http://github.com/robinhood/thorn'
__docformat__ = 'restructuredtext'

# -eof meta-

__all__ = [
    'Thorn', 'Event', 'ModelEvent',
    'Q', 'model_reverser', 'webhook_model',
]



from __future__ import absolute_import, unicode_literals

from thorn import _state

from .case import Case, Mock


class test_current_app(Case):

    def setup(self):
        self.Thorn = self.patch('thorn.app.Thorn')
        self.tls = self.patch('thorn._state._tls')
        self.prev_default_app, _state.default_app = _state.default_app, None

    def teardown(self):
        _state.default_app = self.prev_default_app

    def test_when_default(self):
        self.tls.current_app = None
        self.assertIs(_state.current_app(), self.Thorn.return_value)
        self.Thorn.assert_called_once_with()

    def test_when_current(self):
        app = Mock(name='current_app')
        _state.set_current_app(app)
        self.assertIs(_state.current_app(), app)

    def test_when_default_set(self):
        default_app = Mock(name='default_app')
        self.tls.current_app = None
        _state.set_default_app(default_app)
        self.assertIs(_state.current_app(), default_app)



from __future__ import absolute_import, unicode_literals

from django.core.exceptions import ObjectDoesNotExist

from thorn.django import signals
from thorn.django.utils import serialize_model

from thorn.tests.case import Mock, SignalCase


class test_signal_dispatcher(SignalCase):

    def test_setup_signals(self):
        x = signals.signal_dispatcher(Mock())
        self.assertDictEqual(x.setup_signals(), {})


class SignalDispatcherCase(SignalCase):
    dispatcher = None
    dispatcher_args = ()

    def setup(self):
        self.fun = Mock(name='fun')
        self.dispatch = self.dispatcher(self.fun, *self.dispatcher_args)
        self.instance = Mock(name='instance')
        self.related_model = Mock(name='related_model')
        self.related_model._meta.app_label = 'app_label'
        self.related_model._meta.model_name = 'model_name'


class test_dispatch_on_create(SignalDispatcherCase):
    dispatcher = signals.dispatch_on_create

    def test_dispatch__when_created(self):
        self.dispatch(self.instance, raw=False, created=True)
        self.fun.assert_called_with(
            self.instance,
            context={'instance': self.instance.pk},
        )

    def test_dispatch__when_not_created(self):
        self.dispatch(self.instance, raw=False, created=False)
        self.fun.assert_not_called()

    def test_dispatch__when_raw(self):
        self.dispatch(self.instance, raw=True, created=True)
        self.fun.assert_not_called()

    def test_connect(self):
        self.dispatch.connect(sender=self.Model)
        self.post_save.connect.assert_called_with(
            self.dispatch, sender=self.Model, weak=False,
        )


class test_dispatch_on_change(SignalDispatcherCase):
    dispatcher = signals.dispatch_on_change

    def test_dispatch__when_created(self):
        self.dispatch(self.instance, raw=False, created=True)
        self.fun.assert_not_called()

    def test_dispatch__when_not_created(self):
        self.dispatch(self.instance, raw=False, created=False)
        self.fun.assert_called_with(
            self.instance,
            context={'instance': self.instance.pk},
        )

    def test_dispatch__when_raw(self):
        self.dispatch(self.instance, raw=True, created=True)
        self.fun.assert_not_called()

    def test_on_pre_save(self):
        self.dispatch.use_transitions = True
        instance = Mock(Name='instance')
        sender = Mock(name='sender')
        self.dispatch.on_pre_save(instance, sender, raw=False)
        sender.objects.get.assert_called_once_with(pk=instance.pk)
        self.assertIs(instance._previous_version, sender.objects.get())

    def test_on_pre_save__ObjectDoesNotExist(self):
        self.dispatch.use_transitions = True
        instance = Mock(Name='instance')
        instance._previous_version = None
        sender = Mock(name='sender')
        sender.objects.get.side_effect = ObjectDoesNotExist()
        self.dispatch.on_pre_save(instance, sender, raw=False)
        sender.objects.get.assert_called_once_with(pk=instance.pk)
        self.assertIsNone(instance._previous_version)

    def test_on_pre_save__disabled(self):
        self.dispatch.use_transitions = False
        instance = Mock(name='instance')
        sender = Mock(name='sender')
        self.dispatch.on_pre_save(instance, sender, raw=False)
        sender.objects.get.assert_not_called()

    def test_on_pre_save__disabled_on_raw(self):
        self.dispatch.use_transitions = True
        instance = Mock(Name='instance')
        sender = Mock(name='sender')
        self.dispatch.on_pre_save(instance, sender, raw=True)
        sender.objects.get.assert_not_called()

    def test_on_pre_save__disabled_on_object_creation(self):
        self.dispatch.use_transitions = True
        instance = Mock(Name='instance')
        instance.pk = None
        sender = Mock(name='sender')
        self.dispatch.on_pre_save(instance, sender, raw=False)
        sender.objects.get.assert_not_called()

    def test_connect(self):
        self.dispatch.connect(sender=self.Model)
        self.post_save.connect.assert_called_with(
            self.dispatch, sender=self.Model, weak=False,
        )


class test_dispatch_on_delete(SignalDispatcherCase):
    dispatcher = signals.dispatch_on_delete

    def test_dispatch(self):
        self.dispatch(self.instance)
        self.fun.assert_called_with(
            self.instance, context={'instance': self.instance.pk})

    def test_connect(self):
        self.dispatch.connect(sender=self.Model)
        self.post_delete.connect.assert_called_with(
            self.dispatch, sender=self.Model, weak=False,
        )


class test_dispatch_on_m2m_change(SignalDispatcherCase):
    dispatcher = signals.dispatch_on_m2m_change
    dispatcher_args = ('tags.sgat',)

    def test_dispatch(self):
        self.dispatch(
            self.instance, pk_set={31, 123}, model=self.related_model,
        )
        self.fun.assert_called_with(self.instance, context={
            'instance': self.instance.pk,
            'model': serialize_model(self.related_model),
            'pk_set': [31, 123],
        })

    def test_connect(self):
        self.dispatch.connect(sender=self.Model)
        self.m2m_changed.connect.assert_called_with(
            self.dispatch.on_m2m_change,
            sender=self.Model.tags.sgat.through, weak=False,
        )

    def test_on_m2m_change__post_add(self):
        handler = self.dispatch.actions['post_add'] = Mock(name='post_add')
        self.dispatch.on_m2m_change(
            sender=self.Model, action='post_add', instance=self.instance,
            model=self.Model, foo=1,
        )
        handler.assert_called_with(
            self.instance, sender=self.Model, model=self.Model, foo=1,
        )

    def test_on_m2m_change__no_handler(self):
        self.dispatch.on_m2m_change(
            sender=self.Model, action='foo',
            instance=self.instance, model=self.Model,
        )


class test_dispatch_on_m2m_add(test_dispatch_on_m2m_change):
    dispatcher = signals.dispatch_on_m2m_add


class test_dispatch_on_m2m_remove(test_dispatch_on_m2m_change):
    dispatcher = signals.dispatch_on_m2m_remove


class test_dispatch_on_m2m_clear(test_dispatch_on_m2m_change):
    dispatcher = signals.dispatch_on_m2m_clear



"""

thorn.django.rest_framework.urls
================================

To include the rest-framework API views in your project use the
:meth:`django.conf.urls.include` function, with a proper namespace argument:

.. code-block:: python

    from django.conf.urls.include import include, url

    urlpatterns = [
        url(r'^hooks/',
            include('thorn.django.rest_framework.urls', namespace='webhook')),
    ]

Two new API endpoints will now be available in your application:

- ``GET /hooks/``

    List of webhook subscriptions related to the currently logged in user.

- ``POST /hooks/``

    Create new subscription owned by the currently logged in user.

- ``GET /hooks/<uuid>/``

    Get detail about specific subscription by unique identifier (uuid).

- ``POST|PATCH /hook/<uuid>/``

    Update subscription given its unique identifier (uuid).

- ``DELETE /hook/<uuid>/``

    Delete subscription given its unique identifier (uuid).

"""
from __future__ import absolute_import, unicode_literals

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^$',
        views.SubscriberList.as_view(), name='list'),
    url(r'^(?P<uuid>[0-9a-fA-F-]+)/$',
        views.SubscriberDetail.as_view(), name='detail'),
]



"""

    thorn.funtests.tasks
    =================================

    Custom Celery worker remote control commands
    used in the Thorn functional test suite.

"""
from __future__ import absolute_import, unicode_literals

from collections import Mapping

from celery.utils.imports import symbol_by_name
from celery.worker.control import Panel

from thorn._state import current_app


@Panel.register
def setenv(state, setting_name, new_value):
    app = current_app()
    value = getattr(app.settings, setting_name, None)
    setattr(app.settings, setting_name, new_value)
    return {'OK': value}


def subscribers_for_event(event):
    app = current_app()
    subs = app.settings.THORN_SUBSCRIBERS.get(event)
    if subs is None:
        subs = app.settings.THORN_SUBSCRIBERS[event] = []
    elif not isinstance(subs, list):
        subs = app.settings.THORN_SUBSCRIBERS[event] = [subs]
    return subs


def find_subscriber(self, subs, url):
    for i, sub in enumerate(subs):
        if sub == url or (isinstance(sub, Mapping) and sub['url'] == url):
            return i


def _unsubscribe(subs, url):
    try:
        subs.pop(find_subscriber(subs, url))
    except TypeError:
        pass


@Panel.register
def hook_subscribe(state, event, url=None, callback=None):
    subs = subscribers_for_event(event)
    _unsubscribe(subs, url)
    subscribers_for_event(event).append(
        symbol_by_name(callback) if callback else url,
    )
    return {'ok': 'url {0!r} subscribed to event {1!r}'.format(url, event)}


@Panel.register
def hook_unsubscribe(state, event, url):
    _unsubscribe(subscribers_for_event(event), url)
    return {'ok': 'url {0!r} unsubscribed from event {1!r}'.format(url, event)}


@Panel.register
def hook_clear(state, event):
    subscribers_for_event(event)[:] = []
    return {'ok': 'removed all subscribers for event {0!r}'.format(event)}
